import random
from typing import Tuple, Optional

import pygame
from pygame import Surface

from core.constants import CellValue, CellValueRanges
from core.logic import Logic
from core.state import World
from tools.vector import vectorDivI
from ui.theme.Theme import Theme
from .GameMode import GameMode
from ..Mouse import Mouse
from ..component.CacheComponent import CacheComponent
from ..component.LayerComponent import LayerComponent


class EditGameMode(GameMode):

    def __init__(self, theme: Theme, world: World):
        super().__init__(theme)
        self.__world = world
        self.__logic = Logic(world)
        self.__mouseButtonDown = False
        self.__layers = [
            CacheComponent(LayerComponent(theme, world, name))
            for name in world.layerNames
        ]
        self.__font = theme.getFont("default")

        # Brush
        self.__brushLayer = "ground"

    # Game loop

    def dispose(self):
        for layer in self.__layers:
            layer.dispose()

    def processInput(self):
        pass

    def update(self):
        self.__logic.executeCommands()

    def render(self, surface: Surface):
        for layer in self.__layers:
            layer.render(surface)
        color = self.theme.getFontColor("default")
        textSurfaces = [
            self.__font.render(message, False, color)
            for message in [
                f"Current brush: {self.__brushLayer}",
                "Left click: add",
                "middle click: fill",
                "right click: remove",
                "F1/F2/F3: Select 'ground'/'impassable'/'objects' layer"
            ]
        ]
        y = surface.get_height()
        for textSurface in reversed(textSurfaces):
            y -= textSurface.get_height() + 1
            surface.blit(textSurface, (0, y))

    # Keyboard handling

    def keyDown(self, key: int):
        if key == pygame.K_F1:
            self.__brushLayer = "ground"
        elif key == pygame.K_F2:
            self.__brushLayer = "impassable"
        elif key == pygame.K_F3:
            self.__brushLayer = "objects"

    # Mouse handling

    def __computeCellCoordintates(self, pixel: Tuple[int, int]) -> Optional[Tuple[int, int]]:
        tileSize = self.theme.getTileset(self.__brushLayer).tileSize
        coords = vectorDivI(pixel, tileSize)
        if not self.__world.contains(coords):
            return None
        return coords

    def __updateCell(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1 or mouse.button2:
            if self.__brushLayer == "ground":
                value = CellValue.GROUND_EARTH
            else:
                minValue = CellValueRanges[self.__brushLayer][0]
                maxValue = CellValueRanges[self.__brushLayer][1]
                value = CellValue(random.randint(minValue, maxValue - 1))
            fill = mouse.button2
        elif mouse.button3:
            if self.__brushLayer == "ground":
                value = CellValue.GROUND_SEA
            else:
                value = CellValue.NONE
            fill = False
        else:
            return
        Command = self.__logic.getSetLayerValueCommand(self.__brushLayer)
        command = Command(cell, value, fill)
        self.__logic.addCommand(command)

    def mouseButtonDown(self, mouse: Mouse):
        coords = self.__computeCellCoordintates(mouse.pixel)
        if coords is None:
            return
        self.__mouseButtonDown = True
        self.__updateCell(coords, mouse)

    def mouseMove(self, mouse: Mouse):
        if not self.__mouseButtonDown:
            return
        coords = self.__computeCellCoordintates(mouse.pixel)
        if coords is None:
            return
        self.__updateCell(coords, mouse)

    def mouseButtonUp(self, mouse: Mouse):
        self.__mouseButtonDown = False

    def mouseEnter(self, mouse: Mouse):
        self.__mouseButtonDown = False

    def mouseLeave(self):
        self.__mouseButtonDown = False
